
importPackage(com.change_vision.jude.api.inf.editor);
importPackage(com.change_vision.jude.api.inf.presentation);
importPackage(com.change_vision.jude.api.inf.model);
importPackage(com.change_vision.jude.api.inf.project);
importPackage(java.util);
importPackage(java.awt.geom);

var X_OFFSET = 10.0;
var Y_OFFSET = 50.0;
var X_CONTEXT_OFFSET = 20.0;
var Y_CONTEXT_OFFSET = 8.0;
var LINE_SHAPE = 'curve_right_angle'; //'line','curve_right_angle',..

run();

function run() {
    var nodes = getGsnNodesOnDiagram();
    if (nodes.length === 0) return;

    var roots = getTreeRoots(nodes);
    TransactionManager.beginTransaction();
    layout(roots);
    updateLinkShape();
    TransactionManager.endTransaction();
}

function layout(roots) {
    roots.forEach(function(root) {
        layoutChildren(root, null, new HashSet(), 0, 0.0);
        adjustMultiIncommingNodes(root, new HashSet());
    });
}

function layoutChildren(node, parentNode, traversedNodes, depthFromRoot, maxXOfLeftBranch) {
    traversedNodes.add(node);

    updateYOfNode(node, parentNode);

    var maxXOfThisBranch = maxXOfLeftBranch;
    var children = getChildren(node);
    children.forEach(function(child) {
        if (traversedNodes.contains(child)) return;
        maxXOfThisBranch = layoutChildren(child, node, traversedNodes,
                                depthFromRoot + 1, maxXOfThisBranch);
    });

    updateXOfNode(node, children, maxXOfLeftBranch);
    maxXOfThisBranch = layoutContextNodes(node, traversedNodes, maxXOfThisBranch);

    return Math.max(maxXOfThisBranch, node.location.x + node.width);
}

function updateYOfNode(node, parentNode) {
    var y = 0.0;
    if (parentNode != null) {
        var maxYOfParent = parentNode.location.y + parentNode.height;
        y = maxYOfParent + Y_OFFSET;
    }
    setLocation(node, node.location.x, y);
}

function updateXOfNode(node, children, maxXOfLeftBranch) {
    if (children.length === 0) {
        var x = maxXOfLeftBranch + X_OFFSET;
        setLocation(node, x, node.location.y);
    } else {
        var x = getCenterXOf(children) - node.width / 2.0;
        setLocation(node, x, node.location.y);
        if (x < maxXOfLeftBranch + X_OFFSET) {
            var deltaX = maxXOfLeftBranch + X_OFFSET - x;
            moveChildren(node, deltaX, 0.0, new HashSet());
        }
    }
}

function layoutContextNodes(node, traversedNodes, maxXOfThisBranch) {
    var contexts = getContexts(node).filter(function(context) {
        return !traversedNodes.contains(context);
    });
    if (contexts.length > 0) {
        var contextTotalHeight = 0.0;
        contexts.forEach(function(context) {
            contextTotalHeight += context.height + Y_CONTEXT_OFFSET;
        });
        contextTotalHeight -= Y_CONTEXT_OFFSET;
        var yy = node.location.y + node.height / 2.0 - contextTotalHeight / 2.0;
        contexts.forEach(function(context) {
            var xx = node.location.x + node.width + X_CONTEXT_OFFSET;
            setLocation(context, xx, yy);
            yy += context.height + Y_CONTEXT_OFFSET;
            maxXOfThisBranch = Math.max(maxXOfThisBranch, xx + context.width);
            traversedNodes.add(context);
        });
    }
    return maxXOfThisBranch;
}

function adjustMultiIncommingNodes(node, traversedNodes) {
    traversedNodes.add(node);
    var sources = getSourcesOf(node);
    if (sources.length > 1) {
        //xをsourceの左右中央にずらす
        var newX = getCenterXOf(sources) - node.width / 2.0;
        var deltaX = newX - node.location.x;
        //yを一番深い場所にずらす
        var deepestSource = getDeepestNode(sources);
        var newY = deepestSource.location.y + deepestSource.height + Y_OFFSET;
        var deltaY = newY - node.location.y;
        moveChildren(node, deltaX, deltaY, new HashSet());
    }

    getChildren(node).forEach(function(child) {
        if (traversedNodes.contains(child)) return;
        adjustMultiIncommingNodes(child, traversedNodes);
    });
}

function updateLinkShape(node) {
    var supportedByLinks = getAllSupportedByLinks(node);
    supportedByLinks.forEach(function(link) {
        link.setProperty('line.shape', LINE_SHAPE);
        var rect1 = link.source.rectangle;
        var rect2 = link.target.rectangle;
        link.setAllPoints(
           [new Point2D.Double(rect1.centerX, rect1.centerY),
            new Point2D.Double(rect1.centerX, (rect1.minY + rect2.maxY)/2.0),
            new Point2D.Double(rect2.centerX, (rect1.minY + rect2.maxY)/2.0),
            new Point2D.Double(rect2.centerX, rect2.centerY)]
           );
    });

    var inContextOfLinks = getAllInContextOfLinks(node);
    inContextOfLinks.forEach(function(link) {
        link.setProperty('line.shape', LINE_SHAPE);
        var rect1 = link.source.rectangle;
        var rect2 = link.target.rectangle;
        var middleX = (rect1.minX + rect2.maxX) / 2.0;
        if (rect1.minX < rect2.minX) {
            middleX = (rect2.minX + rect1.maxX) / 2.0;
        }
        link.setAllPoints(
           [new Point2D.Double(rect1.centerX, rect1.centerY),
            new Point2D.Double(middleX, rect1.centerY),
            new Point2D.Double(middleX, rect2.centerY),
            new Point2D.Double(rect2.centerX, rect2.centerY)]
           );
    });
}

function moveChildren(node, deltaX, deltaY, traversedNodes) {
    traversedNodes.add(node);
    setLocation(node, node.location.x + deltaX,
                      node.location.y + deltaY);
    getChildren(node).forEach(function(child) {
        if (traversedNodes.contains(child)) return;
        moveChildren(child, deltaX, deltaY, traversedNodes);
    });
}

function getDeepestNode(nodes) {
    //TODO maxY rather than maxDepth
    var maxDepth = 0;
    var deepestNode = null;
    nodes.forEach(function(node) {
        var depth = getDepth(node, new HashSet());
        if (depth > maxDepth) {
            deepestNode = node;
            maxDepth = depth;
        }
    });
    return deepestNode;
}

function getDepth(node, traversedNodes) {
    var depth = 0;
    traversedNodes.add(node);
    var parent = getParent(node);
    if (parent != null && !traversedNodes.contains(parent)) {
        depth = getDepth(parent, traversedNodes);
        depth++;
    }
    return depth;
}

function getCenterXOf(nodes) {
    var minX = nodes[0].location.x;
    var maxX = minX;
    nodes.forEach(function(node) {
        minX = Math.min(minX, node.location.x);
        maxX = Math.max(maxX, node.location.x + node.width);
    });
    return (minX + maxX) / 2.0;
}

function getAllPresentationsInDiagramEditor() {
    var diagram = astah.viewManager.diagramViewManager.currentDiagram;
    if (diagram == null) {
        println('Please open a diagram.');
        return [];
    }
    return diagram.presentations;
}

function getTreeRoots(presentations) {
    var roots = new HashSet();
    presentations.forEach(function(p) {
        var root = getTreeRoot(p, new HashSet());
        if (root.parent == null) { //ignore contents of modules
            roots.add(root);
        }
    });
    var rootsArray = roots.toArray();
    rootsArray.sort(function(a, b) {
        return a.location.x - b.location.x;
    });
    return rootsArray;
}

function getTreeRoot(node, traversedNodes) {
    traversedNodes.add(node);
    var parent = getParent(node);
    if (parent == null) {
        return node;
    } else {
        if (traversedNodes.contains(parent)) return node;
        return getTreeRoot(parent, traversedNodes);
    }
}

function getGsnNodesOnDiagram() {
    var presentations = getAllPresentationsInDiagramEditor();
    if (presentations.length === 0) return [];

    return presentations.filter(function(p) {
        return isGsnNode(p.model);
    });
}

function getAllSupportedByLinks() {
    return getAllPresentationsInDiagramEditor().filter(function(p) {
        return isSupportedBy(p.model);
    });
}
function getAllInContextOfLinks() {
    return getAllPresentationsInDiagramEditor().filter(function(p) {
        return isInContextOf(p.model);
    });
}

function getChildren(node) {
    var children = [];
    node.links.forEach(function(link) {
        if (isSupportedBy(link.model) && link.target == node) {
            children.push(link.source);
        }
    });

    return children.sort(function(a, b) {
        return a.location.x - b.location.x;
    });
}

function getSourcesOf(node) {
    var sources = [];
    node.links.forEach(function(link) {
        if (isSupportedBy(link.model) && link.source == node) {
            sources.push(link.target);
        }
    });
    return sources;
}

/* Return Context, Assumption and Justification */
function getContexts(node) {
    var contexts = [];
    node.links.forEach(function(link) {
        if (isInContextOf(link.model) && link.target == node) {
            contexts.push(link.source);
        }
    });

    return contexts.sort(function(a, b) {
        return a.location.y - b.location.y;
    });
}

function getParent(node) {
    var links = node.links;
    //check supportedBy first
    for (var i = 0; i < links.length; i++) {
        var link = links[i];
        if (isSupportedBy(link.model) && link.source == node) {
            return link.target;
        }
    }
    //check inContextOf next
    for (var i = 0; i < links.length; i++) {
        var link = links[i];
        if (isInContextOf(link.model) && link.source == node) {
            return link.target;
        }
    }
    return null;
}

function isSupportedBy(model) {
    return getStereotype(model) == 'supportedBy';
}

function isInContextOf(model) {
    return getStereotype(model) == 'inContextOf';
}

function isGoal(model) {
    return getStereotype(model) == 'goal';
}

function isStrategy(model) {
    return getStereotype(model) == 'strategy';
}

function isSolution(model) {
    return getStereotype(model) == 'solution';
}

function isModule(model) {
    return getStereotype(model) == 'module';
}

function isContext(model) {
    return getStereotype(model) == 'context';
}

function isAssumption(model) {
    return getStereotype(model) == 'assumption';
}

function isJustification(model) {
    return getStereotype(model) == 'justification';
}

function isGsnNode(model) {
    if (isGoal(model) || isStrategy(model) || isSolution(model)
        || isContext(model) || isJustification(model)
        || isAssumption(model) || isModule(model)) {
        return true;
    }
    return false;
}

function getStereotype(model) {
    if (model == null) return '';
    var stereotypes = model.getStereotypes();
    if (stereotypes != null && stereotypes.length > 0) {
        return stereotypes[0];
    }
    return '';
}

function setLocation(node, x, y) {
    node.setLocation(new Point2D.Double(x, y));
}
